import json
from base64 import urlsafe_b64encode

import fido2.features
from django.conf import settings
from django.http import JsonResponse
from django.utils import timezone
from django.views.decorators.csrf import csrf_exempt
from fido2.server import Fido2Server
from fido2.utils import websafe_decode, websafe_encode
from fido2.webauthn import PublicKeyCredentialRpEntity, AttestedCredentialData, RegistrationResponse
from .models import UserPasskey
from user_agents.parsers import parse as ua_parse


def enable_json_mapping():
    try:
        fido2.features.webauthn_json_mapping.enabled = True
    except:
        pass


def getUserCredentials(user):
    return [AttestedCredentialData(websafe_decode(uk.token)) for uk in UserPasskey.objects.filter(user = user)]


def getServer(request=None):
    """Get Server Info from settings and returns a Fido2Server"""
    if callable(settings.FIDO_SERVER_ID):
        fido_server_id = settings.FIDO_SERVER_ID(request)
    else:
        fido_server_id = settings.FIDO_SERVER_ID

    if callable(settings.FIDO_SERVER_NAME):
        fido_server_name = settings.FIDO_SERVER_NAME(request)
    else:
        fido_server_name = settings.FIDO_SERVER_NAME

    rp = PublicKeyCredentialRpEntity(id=fido_server_id, name=fido_server_name)
    return Fido2Server(rp)


def get_current_platform(request):
    ua = ua_parse(request.META["HTTP_USER_AGENT"])
    if 'Safari' in ua.browser.family:
        return "Apple"
    elif 'Chrome' in ua.browser.family and ua.os.family == "Mac OS X":
        return "Chrome on Apple"
    elif 'Android' in ua.os.family:
        return "Google"
    elif "Windows" in ua.os.family:
        return "Microsoft"
    else: return "Key"


def reg_begin(request):
    """Starts registering a new FIDO Device, called from API"""
    enable_json_mapping()
    server = getServer(request)
    auth_attachment = getattr(settings,'KEY_ATTACHMENT', None)
    registration_data, state = server.register_begin({
        u'id':  urlsafe_b64encode(request.user.username.encode("utf8")),
        u'name': request.user.username,
        u'displayName': request.user.username,
    }, getUserCredentials(request.user), authenticator_attachment = auth_attachment, resident_key_requirement=fido2.webauthn.ResidentKeyRequirement.PREFERRED)
    request.session['fido2_state'] = state
    return JsonResponse(dict(registration_data))
    #return HttpResponse(cbor.encode(registration_data), content_type = 'application/octet-stream')


@csrf_exempt
def reg_complete(request):
    """Completes the registeration, called by API"""
    try:
        if not "fido2_state" in request.session:
            return JsonResponse({'status': 'ERR', "message": "FIDO Status can't be found, please try again"})
        enable_json_mapping()
        data = json.loads(request.body)
        name = data.pop("key_name",'')
        server = getServer(request)
        auth_data = server.register_complete(request.session.pop("fido2_state"), response = data)
        encoded = websafe_encode(auth_data.credential_data)
        platform = get_current_platform(request)
        if name == "":
            name = platform
        uk = UserPasskey(user=request.user, token=encoded, name = name,platform=platform)
        if data.get("id"):
            uk.credential_id = data.get('id')

        uk.save()
        return JsonResponse({'status': 'OK'})
    except Exception as exp:
        import traceback
        print(traceback.format_exc())
        return JsonResponse({'status': 'ERR', "message": "Error on server, please try again later"})


def auth_begin(request):
    enable_json_mapping()
    server = getServer(request)
    credentials=[]
    username = None
    if "base_username" in request.session:
        username = request.session["base_username"]
    if request.user.is_authenticated:
        username = request.user.username
    if username:
        credentials = getUserCredentials(request.session.get("base_username", request.user.username))
    auth_data, state = server.authenticate_begin(credentials)
    request.session['fido2_state'] = state
    return JsonResponse(dict(auth_data))


@csrf_exempt
def auth_complete(request):
    enable_json_mapping()
    credentials = []
    server = getServer(request)
    data = json.loads(request.POST["passkeys"])
    key = None
    #userHandle = data.get("response",{}).get('userHandle')
    credential_id = data['id']
    #
    # if userHandle:
    #     if User_Passkey.objects.filter(=userHandle).exists():
    #         credentials = getUserCredentials(userHandle)
    #         username=userHandle
    #     else:
    #         keys = User_Keys.objects.filter(user_handle = userHandle)
    #         if keys.exists():
    #             credentials = [AttestedCredentialData(websafe_decode(keys[0].properties["device"]))]

    keys = UserPasskey.objects.filter(credential_id = credential_id,enabled=1)
    if keys.exists():

        credentials=[AttestedCredentialData(websafe_decode(keys[0].token))]
        key = keys[0]

        try:
            cred = server.authenticate_complete(
                    request.session.pop('fido2_state'), credentials = credentials, response = data
            )
        except ValueError:
            return None
        except Exception as excep:
            raise Exception(excep)
        if key:
            key.last_used = timezone.now()
            request.session["passkey"] = {'passkey': True, 'name': key.name, "id":key.id, "platform": key.platform,
                                           'cross_platform': get_current_platform(request) != key.platform}
            key.save()
            return key.user
    return None



from django.urls import path
from . import FIDO2,views

app_name = 'passkeys'
urlpatterns = [
    path('auth/begin',FIDO2.auth_begin, name='auth_begin'),
    path('auth/complete',FIDO2.auth_complete, name='auth_complete'),
    path('reg/begin',FIDO2.reg_begin,name = 'reg_begin'),
    path('reg/complete',FIDO2.reg_complete,name = 'reg_complete'),
    path('',views.index, name='home'),
    path('enroll/',views.index, name='enroll',kwargs={'enroll':True}),


    path('del/',views.delKey, name='delKey'),
    path('toggle/',views.toggleKey, name='toggle'),
    ]


from django.contrib.auth.decorators import login_required
from django.http import HttpResponse
from django.shortcuts import render

from .models import UserPasskey

@login_required
def index(request,enroll=False):
    keys = UserPasskey.objects.filter(user=request.user)
    return render(request,'PassKeys.html',{"keys":keys,"enroll":enroll})


@login_required
def delKey(request):
    key=UserPasskey.objects.get(id=request.GET["id"])
    if key.user.pk  == request.user.pk:
        key.delete()
        return HttpResponse("Deleted Successfully")
    return HttpResponse("Error: You own this token so you can't delete it")

@login_required
def toggleKey(request):
    id=request.GET["id"]
    q=UserPasskey.objects.filter(user=request.user, id=id)
    if q.count()==1:
        key=q[0]
        key.enabled=not key.enabled
        key.save()
        return HttpResponse("OK")
    return HttpResponse("Error")



# Generated by Django 2.2.28 on 2022-10-28 13:18

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='UserPasskey',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('enabled', models.BooleanField(default=True)),
                ('platform', models.CharField(default='', max_length=255)),
                ('added_on', models.DateTimeField(auto_now_add=True)),
                ('last_used', models.DateTimeField(default=None, null=True)),
                ('credential_id', models.CharField(max_length=255, unique=True)),
                ('token', models.CharField(max_length=255)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



#!/usr/bin/env python

from setuptools import find_packages, setup

setup(
    name='django-passkeys',
    version='1.2.1',
    description='A Django Authentication Backend for Passkeys',
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author='Mohamed El-Kalioby',
    author_email = 'mkalioby@mkalioby.com',
    url = 'https://github.com/mkalioby/django-passkeys',
    download_url='https://github.com/mkalioby/django-passkeys',
    license='MIT',
    packages=find_packages(),
    install_requires=[
        'django >= 2.0',
        'ua-parser',
        'user-agents',
        'fido2 == 1.1.1',
      ],
    python_requires=">=3.7",
    include_package_data=True,
    zip_safe=False, # because we're including static files
    classifiers=[
        #"Development Status :: 5 - Production/Stable",
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 2.0",
        "Framework :: Django :: 2.1",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.0",
        "Framework :: Django :: 4.1",
        "Framework :: Django :: 4.2",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Software Development :: Libraries :: Python Modules",
]
)



from django.contrib.auth import get_user_model
from django.db import models


class UserPasskey(models.Model):
    user_model = get_user_model()
    user = models.ForeignKey(user_model,on_delete=models.CASCADE)
    name = models.CharField(max_length=255)
    enabled= models.BooleanField(default=True)
    platform = models.CharField(max_length=255,default='')
    added_on = models.DateTimeField(auto_now_add=True)
    last_used = models.DateTimeField(null=True,default=None)
    credential_id = models.CharField(max_length=255, unique=True)
    token = models.CharField(max_length=255, null=False)